package jp.sf.amateras.javalib.mockrunner.jdbc;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;

import jp.sf.amateras.javalib.mockrunner.jdbc.CustomerDao.Customer;

import com.mockrunner.jdbc.BasicJDBCTestCaseAdapter;
import com.mockrunner.jdbc.FileResultSetFactory;
import com.mockrunner.jdbc.PreparedStatementResultSetHandler;
import com.mockrunner.jdbc.XMLResultSetFactory;
import com.mockrunner.mock.jdbc.MockResultSet;

public class CustomerDaoTest extends BasicJDBCTestCaseAdapter {

	@Override
	public void setUp() throws Exception {
		super.setUp();
		// 正規表現を有効にする
		setUseRegularExpressions(true);
	}

	public void testSelect() throws SQLException {
		int id = 1;
		String firstname = "山田";
		String lastname = "花子";
		String street = null;
		String city = "東京";

		PreparedStatementResultSetHandler handler =
			getPreparedStatementResultSetHandler();
		// モックのResultSet生成
		MockResultSet mockResultSet = handler.createResultSet();
		// モックが返すResultSetの値を設定する
		mockResultSet.addRow(new Object[]{id, firstname, lastname, street, city});
		handler.prepareGlobalResultSet(mockResultSet);

		// 1件取得
		CustomerDao dao = new CustomerDao();
		Customer customer = dao.selectByKey(1);

		assertEquals(id, customer.getId());
		assertEquals(firstname, customer.getFirstname());
		assertEquals(lastname, customer.getLastname());
		assertEquals(street, customer.getStreet());
		assertEquals(city, customer.getCity());

		// 期待通りのSQLが発行されたか確認
		String sql = "SELECT ID, FIRSTNAME, LASTNAME, STREET, CITY FROM CUSTOMER WHERE ID = \\?";
		verifySQLStatementParameter(sql, 0, 1, id);
		verifyAllResultSetsClosed();
		verifyAllStatementsClosed();
	}

	public void testSelectFile() throws SQLException {
		FileResultSetFactory factory = new FileResultSetFactory(
				"jp/sf/amateras/javalib/mockrunner/jdbc/Customer.csv");
		// 1行目はカラム名であることを設定
		factory.setFirstLineContainsColumnNames(true);
		// デリミタを指定
		factory.setDelimiter(",");
		// デフォルトのテンプレートを利用する
		factory.setUseTemplates(true);

		// モックのResultSet生成
		String id = "customerResultSet";
		MockResultSet mockResultSet = factory.create(id);

		String sql = "SELECT .* FROM CUSTOMER WHERE ID = \\?";
		getPreparedStatementResultSetHandler().setUseRegularExpressions(true);
		getPreparedStatementResultSetHandler().prepareResultSet(sql, mockResultSet);

		// 1件取得
		CustomerDao dao = new CustomerDao();
		int key = 1;
		Customer customer = dao.selectByKey(key);

		assertNotNull(customer);

		// 期待通りのSQLが発行されたか確認
		verifySQLStatementParameter(sql, 0, 1, key);
		verifyResultSetClosed(id);
		verifyAllStatementsClosed();
	}

	public void testSelectFileCustom() throws SQLException {
		FileResultSetFactory factory = new FileResultSetFactory(
				"jp/sf/amateras/javalib/mockrunner/jdbc/Customer_custom.csv");
		// 1行目はカラム名であることを設定
		factory.setFirstLineContainsColumnNames(true);

		// カスタムテンプレート作成
		Map<String, String> templates = new HashMap<String, String>();
		templates.put("defaultCity", "東京");
		// カスタムテンプレートを利用する
		factory.setTemplateConfiguration("$", templates);
		factory.setUseTemplates(true);

		// モックのResultSet生成
		MockResultSet mockResultSet =
			getPreparedStatementResultSetHandler().createResultSet(factory);
		getPreparedStatementResultSetHandler().prepareGlobalResultSet(mockResultSet);

		// 1件取得
		CustomerDao dao = new CustomerDao();
		Customer customer = dao.selectByKey(1);

		assertEquals("東京", customer.getCity());
	}

	public void testSelectXmlFile() throws SQLException {
		XMLResultSetFactory factory = new XMLResultSetFactory(
				"jp/sf/amateras/javalib/mockrunner/jdbc/Customer.xml");

		// モックのResultSet生成
		MockResultSet mockResultSet =
			getPreparedStatementResultSetHandler().createResultSet(factory);
		getPreparedStatementResultSetHandler().prepareGlobalResultSet(mockResultSet);

		// 1件取得
		CustomerDao dao = new CustomerDao();
		Customer customer = dao.selectByKey(1);

		assertEquals(1, customer.getId());
		assertEquals("山田", customer.getFirstname());
		assertEquals("花子", customer.getLastname());
		assertEquals("", customer.getStreet());
		assertEquals("東京", customer.getCity());
	}

	public void testUpdate() throws SQLException {
		Customer customer = new Customer();
		customer.setId(100);
		customer.setFirstname("山田");
		customer.setLastname("太郎");
		customer.setCity("大阪");

		// 1件更新
		CustomerDao dao = new CustomerDao();
		dao.updateByKey(customer);

		// 期待通りのSQLが発行されたか確認
		String sql = "UPDATE CUSTOMER .* WHERE ID = \\?";
		verifySQLStatementParameter(sql, 0, 1, customer.getFirstname());
		verifySQLStatementParameter(sql, 0, 2, customer.getLastname());
		verifySQLStatementParameter(sql, 0, 3, customer.getStreet());
		verifySQLStatementParameter(sql, 0, 4, customer.getCity());
		verifySQLStatementParameter(sql, 0, 5, customer.getId());
		verifyCommitted();
		verifyNotRolledBack();
		verifyPreparedStatementClosed(sql);
	}

	public void testUpdateException() throws SQLException {
		Customer customer = new Customer();
		customer.setId(100);
		customer.setFirstname("山田");
		customer.setLastname("太郎");
		customer.setCity("大阪");

		PreparedStatementResultSetHandler handler =
			getPreparedStatementResultSetHandler();
		// 期待するSQL
		String sql = "UPDATE CUSTOMER .* WHERE ID = \\?";
		// ハンドラに対して正規表現を有効にする
		handler.setUseRegularExpressions(true);
		// 例外を発生させるように設定する
		handler.prepareThrowsSQLException(sql);

		CustomerDao dao = new CustomerDao();
		try {
			// 1件更新
			dao.updateByKey(customer);
			fail();

		} catch (Exception e) {
			assertTrue(e instanceof SQLException);
		}

		// ロールバックされたか確認
		verifyNotCommitted();
		verifyRolledBack();
		verifyPreparedStatementClosed(sql);
	}
}
